/*
 * File: isearch.c
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.  This file was extended
 * by D.R. Banks, John M. Gamble, and Dan Corkill.
 *
 */


/***

The functions in this file implement commands that perform incremental
searches in the forward and backward directions. Contains references
to routines internal to search.c.

***/

#include	<ctype.h>
#include	<stdio.h>
#include	<string.h>
#include	"rho.h"
#include	"isearch.h"
#include	"bind.h"
#include	"cursor.h"
#include	"display.h"
#include	"file.h"
#include	"macros.h"
#include	"messages.h"
#include	"search.h"
#include	"utils.h"



#define CMDBUFLEN	256     /* Length of our command buffer */


static int (*saved_get_char) ();        /* Get character routine */
static int cmd_buff[CMDBUFLEN]; /* Save the command args here */
static int cmd_offset;          /* Current offset into command buff */
static int cmd_reexecute = -1;  /* > 0 if re-executing command */
static int eaten_char = -1;     /* Re-eaten char */





/*++

Search for the next occurrence of <pat> in the buffer, either forward
or backward.  If we can't find any more matches, "point" is left where
it was before.  If we do find a match, "point" will be at the end of the
matched string for forward searches and at the beginning of the matched
string for reverse searches.

++*/
static int scanmore(dir)
int dir;                        /* direction to search          */
{
    int sts;                    /* search status                */

    setjtable();                /* Set up fast search arrays    */

    sts = scanner(dir, (dir == REVERSE) ? PTBEG : PTEND, 1);

    if (!sts)
        TTbeep();               /* Feep if search fails       */

    return (sts);
}

/*++

Trivial routine to insure that the next character in the search string
is still true to whatever we're pointing to in the buffer.  This routine
will not attempt to move the "point" if the match fails, although it
will implicitly move the "point" if we're forward searching, and find
a match, since that's the way forward isearch works.  If we are reverse
searching we compare all characters in the pattern string from "point"
to the new end.  If the compare fails, we return FALSE and call scanmore.

++*/
static int checknext(chr, dir)
int chr;                        /* Next char to look for        */
int dir;                        /* Search direction             */
{
    LINE *curline;              /* current line during scan     */
    int curoff;                 /* position within current line */
    char *patrn;                /* The entire search string (incl chr) */
    int sts;                    /* how well things go           */

    /* setup the local scan pointer to current "." */

    curline = curwp->w_dotp;    /* Get the current line structure */
    curoff = curwp->w_doto;     /* Get the offset within that line */

    if (dir == FORWARD) {       /* If searching forward         */
        if ((sts = !boundary(curline, curoff, FORWARD))) {
            /* If it's what we're looking for, set the point
             * and say that we've moved.
             */
            if ((sts =
                 eq(nextch(&curline, &curoff, FORWARD),
                    (unsigned char) chr))) {
                curwp->w_dotp = curline;
                curwp->w_doto = curoff;
                curwp->w_flag |= WFMOVE;
            }
        }
    } else {                    /* Else, reverse search check. */
        patrn = (char *) pat;
        while (*patrn) {        /* Loop for all characters in patrn   */
            if ((sts = !boundary(curline, curoff, FORWARD)) == FALSE ||
                (sts =
                 eq(nextch(&curline, &curoff, FORWARD),
                    (unsigned char) *patrn)) == FALSE)
                break;          /* Nope, just punt it then */
            patrn++;
        }
    }

    /*
     * If the 'next' character didn't fit in the pattern,
     * let's go search for it somewhere else.
     */
    if (sts == FALSE)
        sts = scanmore(dir);

    return (sts);               /* And return the status */
}


/*++

Routine to get the next character or extended character from the input
stream.  If we're reading from the real terminal, force a screen update
before we get the char.  Otherwise, we must be re-executing the command
string, so just return the next character.

++*/
static int get_char()
{
    int c;
    KEYTAB *key;

    /* See if we're re-executing: */
    if (cmd_reexecute >= 0)     /* Is there an offset?          */
        if ((c = cmd_buff[cmd_reexecute++]) != 0)
            return (c);         /* Yes, return any character    */

    /* We're not re-executing (or aren't any more).  Try for a real char */
    cmd_reexecute = -1;         /* Say we're in real mode again */
    update(FALSE);              /* Pretty up the screen         */
    if (posflag)
        update_modeline();
    if (cmd_offset >= CMDBUFLEN - 1) {  /* If we're getting too big ... */
        mlwrite(TEXT167);       /* "? command too long" */
        return (isterm);        /* force quit           */
    }

    /*
     * If $isterm != meta character, then create the extended
     * character as get_keystrokes() does.
     */
    if ((c = mlgetkey()) == (int) isterm)
        return (isterm);
    if ((key = exist_binding(c)) != NULL) {
        if (key->k_fp == ctlx || key->k_fp == meta_prefix) {
            c = mlgetkey();
            c = toupper(c) | (c & ~255);        /* Force to upper */
            c |= (key->k_fp == ctlx) ? CTLX : META;
        }
    }

    cmd_buff[cmd_offset++] = (int) c;   /* Save the char for next time  */
    cmd_buff[cmd_offset] = '\0';        /* And terminate the buffer     */
    return (c);                 /* Return the character         */
}



/*++

Routine to un-eat a character.  This will save the character to be
re-eaten by redirecting the input call to a routine here.  Come here on
the next term.t_getchar call.

++*/
static int uneat()
{
    int c;

    term.t_getchar = saved_get_char;    /* restore the routine address  */
    c = eaten_char;             /* Get the re-eaten char        */
    eaten_char = -1;            /* Clear the old char           */
    return (c);                 /* and return the last char     */
}




/*++

Routine to re-eat a character.

++*/
static void reeat(c)
int c;
{
    if (eaten_char != -1)       /* If we've already been here   */
        return;                 /* Don't do it again            */
    eaten_char = c;             /* Else, save the char for later */
    saved_get_char = term.t_getchar;    /* Save the char get routine    */
    term.t_getchar = uneat;     /* Replace it with ours         */
}



/*++

Subroutine to do an incremental search.

While searching forward, each successive character will leave the cursor
at the end of the entire matched string.  Typing a Control-S will cause
the next occurrence of the string to be searched for (where the next
occurrence does NOT overlap the current occurrence).  A Control-R will
change to a backwards search, ESC will terminate the search and Control-G
will abort the search.  BKSP will back up to the previous match of the
string, or if the starting point is reached first, it will delete the
last character from the search string.

While searching backward, each successive character will leave the cursor
at the beginning of the matched string.  Typing a Control-R will search
backward for the next occurrence of the string.  Control-S will revert
the search to the forward direction.  In general, the reverse incremental
search is just like the forward incremental search inverted.

In all cases, if the search fails, the user will be beeped, and the search
will stall until the pattern string is edited back into something that
exists or is aborted.

++*/
static int isearch(dir)
int dir;

{

    char pat_save[NPAT];        /* Saved copy of the old pattern str  */
    int (*kfunc) ();            /* ptr to the requested function to bind to */
    int c;                      /* current input character */
    int col;                    /* prompt column */
    int cpos;                   /* character number in search string  */
    int curoff;                 /* Current offset on entry                */
    int expc;                   /* function expanded input char           */
    int init_direction;         /* The initial search direction           */
    int status;                 /* Search status */
    KEYTAB *ktp;                /* The command bound to the key           */
    LINE *curline;              /* Current line on entry                  */

    /* Set up the starting conditions */

    cmd_reexecute = -1;         /* We're not re-executing (yet?)          */
    cmd_offset = 0;             /* Start at the beginning of cmd_buff */
    cmd_buff[0] = '\0';         /* Reset the command buffer       */
    strncpy(pat_save, (char *) pat, NPAT);      /* Save the old pattern string            */
    curline = curwp->w_dotp;    /* Save the current line pointer          */
    curoff = curwp->w_doto;     /* Save the current offset                */
    init_direction = dir;       /* Save the initial search direction  */


  start_over:

    /*
     * Ask the user for the text of a pattern,
     * and remember the column.
     */
    col = (clexec) ? 0 : mlprompt(TEXT165, (char *) pat, isterm);       /*"ISearch: " */

    cpos = 0;                   /* Start fresh */
    status = TRUE;

    for (;;) {                  /* ISearch per character loop */

        /* Check for special characters first.
           That is, a control or ^X or FN press.
           Most cases here change the search.
         */
        c = ectoc(expc = get_char());

        if (expc == isterm) {   /* Want to quit searching?        */
            setjtable();        /* Update jump tables...          */
            return (TRUE);      /* Quit searching now             */
        }

        if (expc == abortc)     /* If abort search request        */
            break;              /* Quit searching                 */

        if (expc == quotec) {   /* Quote character?       */
            c = ectoc(expc = get_char());       /* Get the next char              */
        } else if ((expc > 255 || expc == 0) && (c != '\t' && c != '\r')) {

            kfunc =
                ((ktp = exist_binding(expc)) == NULL) ? NULL : ktp->k_fp;

            if (kfunc == fisearch || kfunc == risearch) {
                dir = (kfunc == risearch) ? REVERSE : FORWARD;

                /*
                 * if cpos == 0 then we are either just starting
                 * or starting over.  Use the original pattern
                 * in pat, which has either not been changed or
                 * has just been restored. Find the length and
                 * re-echo the string.
                 */
                if (cpos == 0) {
                    cpos = (int) strlen((char *) pat);
                    col = echostring((char *) pat, col, NPAT / 2);
                }

                status = scanmore(dir);
                continue;
            } else if (kfunc == delete_backward) {
                if (cmd_offset <= 1) {  /* Anything to delete?            */
                    return (TRUE);      /* No, just exit                  */
                }

                /* Back up over the Rubout and the character
                 * it's rubbing out.  If it is a quoted
                 * char, rub the quote char out too.
                 */
                cmd_offset -= 2;
                if (cmd_offset > 0
                    && cmd_buff[cmd_offset - 1] == (int) quotec)
                    cmd_offset--;
                cmd_buff[cmd_offset] = '\0';

                curwp->w_dotp = curline;        /* Reset the line pointer         */
                curwp->w_doto = curoff; /*      and the offset            */
                dir = init_direction;   /* Reset the search direction */
                strncpy((char *) pat, pat_save, NPAT);  /* Restore the old search str */
                setjtable();    /* and its jump tables.           */
                cmd_reexecute = 0;      /* Start the whole mess over  */
                goto start_over;        /* Let it take care of itself */
            }

            /* Presumably the key was uninteresting... */

            reeat(expc);        /* Re-eat the char                */
            return (TRUE);      /* And return the last status */
        }

        /* I guess we got something to search for, so search for it       */
        pat[cpos++] = (unsigned char) c;        /* put the char in the buffer */

        if (cpos >= NPAT) {     /* too many chars in string?  */
            mlwrite(TEXT166);   /* "? Search string too long" */
            strncpy((char *) pat, pat_save, NPAT);      /* Restore old search str */
            setjtable();        /* and its jump tables.           */
            return (FALSE);     /* Return an error, but stay. */
        }

        pat[cpos] = 0;          /* null terminate the buffer  */
        col = echo_char(c, col);        /* Echo the character             */
        if (!status)            /* If we lost last time           */
            TTbeep();           /* beep again           */
        else                    /* Otherwise, we must have won */
            status = checknext(c, dir); /* See if still matches or find next */

    }

    curwp->w_dotp = curline;    /* Reset the line pointer               */
    curwp->w_doto = curoff;     /*   and the offset to original value   */
    curwp->w_flag |= WFMOVE;    /* Say we've moved                      */
    update(FALSE);              /* And force an update                  */
    return (FALSE);
}



/*++

Subroutine to do incremental reverse search.  It actually uses the
same code as the normal incremental search, as both can go both ways.

++*/
int risearch(f, n)
int f, n;                       /* prefix flag and argument */
{
    f = n = 0;                  /* splint */

    return fisearch(f, 2);
}

/*++

Subroutine to do incremental forward search.  It actually uses the
same code as the reverse incremental search, as both can go both ways.

++*/
int fisearch(f, n)
int f, n;
{
    int status;
    int direction;

    f = 0;                      /* splint */
    if (n == 2) {
        /* Make sure search doesn't match where we are by backing up a character. */
        previous_character(TRUE, 1);
        direction = REVERSE;
    } else {
        direction = FORWARD;
    }

    if ((status = isearch(direction)))
        mlerase();
    else
        mlwrite(TEXT164);       /* "[search failed]" */
    return (status);
}
